/*
---------------------------------------------------------------------------
 $Id: config.c,v 1.51 2006/09/22 23:06:41 cnepveu Exp $
---------------------------------------------------------------------------
Copyright (c) 2001-2006 Hexago Inc. All rights reserved.

  LICENSE NOTICE: You may use and modify this source code only if you
  have executed a valid license agreement with Hexago Inc. granting
  you the right to do so, the said license agreement governing such
  use and modifications.   Copyright or other intellectual property
  notices are not to be removed from the source code.
---------------------------------------------------------------------------
*/

/*  Configuration file handling. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include <sys/types.h>

#define _USES_NETDB_H_
#define _USES_SYS_IOCTL_H_
#define _USES_SYS_SOCKET_H_
#define _USES_NETINET_IN_H_
#define _USES_NET_IF_H_
#define _USES_ARPA_INET_H_
#define _USES_SYSLOG_H_

#include "platform.h"

#include "config.h"
#include "tsp_client.h"	// tspGetLocalAddress
#include "log.h"
#include "hex_strings.h"
#include "lib.h"
#include "cli.h"
#include "version.h"
#include "net.h"

/* A flag to remember if the configuration file contains a 'log' directive */
/* If it doesn't, we need to apply default values. */
static int log_directive_parsed = 0;

static syslog_facility_t syslog_facilities[] = {
	{ STR_CONFIG_SLOG_FACILITY_USER, LOG_USER },
	{ STR_CONFIG_SLOG_FACILITY_LOCAL0, LOG_LOCAL0 },
	{ STR_CONFIG_SLOG_FACILITY_LOCAL1, LOG_LOCAL1 },
	{ STR_CONFIG_SLOG_FACILITY_LOCAL2, LOG_LOCAL2 },
	{ STR_CONFIG_SLOG_FACILITY_LOCAL3, LOG_LOCAL3 },
	{ STR_CONFIG_SLOG_FACILITY_LOCAL4, LOG_LOCAL4 },
	{ STR_CONFIG_SLOG_FACILITY_LOCAL5, LOG_LOCAL5 },
	{ STR_CONFIG_SLOG_FACILITY_LOCAL6, LOG_LOCAL6 },
	{ STR_CONFIG_SLOG_FACILITY_LOCAL7, LOG_LOCAL7 },
	{ NULL, 0 }
};

/* Verify dns server for dns delegation */
static
int
VerifyDnsServer(char* server)
{
	char buffer[INET6_ADDRSTRLEN];

	/* if its neither a valid v4 nor v6 address, never mind */
	
	if( INET_PTON(AF_INET, server, buffer) == 0 && INET_PTON(AF_INET6, server, buffer) == 0 )
		return 1; 

	return 0;
}

/* Parse the configuration file's 'log' directive */
static
int
ParseLogDirective(tConf *conf, char *log)
{
	char *destination, *level;
	int numeric_level;

	/* The 'log=' part has already been consumed when we get here. */
	/* Get the first part of what remains, ie the 'destination' part in */
	/* 'destination;level'. */
	destination = strtok(log, ";");

	if (destination != NULL) {
		/* Get the second ('level') part. */
		level = strtok(NULL, ";");

		if (level != NULL) {
			/* If level is specified, turn it into a number, and use it. */
			numeric_level = atoi(level);
			
			if ((numeric_level < LOG_LEVEL_DISABLED) || (numeric_level > LOG_LEVEL_MAX)) {
				Display(LOG_LEVEL_1, ELError, "ParseLogDirective", HEX_STR_LOG_LEVEL_OUT_OF_BOUNDS, log, numeric_level);
				return 1;
			}
		}
		else {
			/* If level is not specified, we use level 1 as the default value. */
			numeric_level = LOG_LEVEL_1;
		}

		if (strcmp(destination, STR_CONFIG_LOG_DESTINATION_STDERR) == 0) {
			/* log=stderr;level */
			conf->log_level_stderr = numeric_level;
		}
		else if (strcmp(destination, STR_CONFIG_LOG_DESTINATION_SYSLOG) == 0) {
			/* log=syslog;level */

#ifdef WIN32
			// Display(LOG_LEVEL_2, ELWarning, "ParseLogDirective", HEX_STR_LOG_SYSLOG_NOT_SUPPORTED);
			return 0;
#endif
			conf->log_level_syslog = numeric_level;
		}
		else if (strcmp(destination, STR_CONFIG_LOG_DESTINATION_CONSOLE) == 0) {
			/* log=console;level */
			conf->log_level_console = numeric_level;
		}
		else if (strcmp(destination, STR_CONFIG_LOG_DESTINATION_FILE) == 0) {
			/* log=file;level */
			conf->log_level_file = numeric_level;
		}
		else {
			/* We don't know about this destination. */
			Display(LOG_LEVEL_1, ELError, "ParseLogDirective", HEX_STR_INVALID_VAL_FOR_LOG, log);
			return 1;
		}		
	}
	else {
		/* Couldn't find the 'destination' part. That's just wrong. */
		Display(LOG_LEVEL_1, ELError, "ParseLogDirective", HEX_STR_INVALID_VAL_FOR_LOG, log);
		return 1;
	}


	/* Remember that we found a log directive. */
	log_directive_parsed = 1;
	return 0;
}

/* Parse the configuration file's 'syslog_facility' directive. */
static
int
ParseSyslogFacility(tConf *conf, char *facility)
{
	int index = 0;

#ifdef WIN32
	// Display(LOG_LEVEL_2, ELWarning, "ParseSyslogFacility", HEX_STR_SYSLOG_FACILITY_NOT_SUPPORTED);
	return 0;
#endif

	/* Loop through the known facility strings, and compare with the one we found. */
	while ((syslog_facilities != NULL) && (syslog_facilities[index].string != NULL)) {
		if (strcmp(facility, syslog_facilities[index].string) == 0) {
			conf->syslog_facility = syslog_facilities[index].value;
			
			return 0;
		}

		index++;
	}

	return 1;
}


/* Add option to configuration struct */
static
int
AddOption(char *Option, char *Value, tConf *Conf)
{
  if (strcmp(Option, "gw6_dir") == 0) {
    Conf->tsp_dir = strdup(Value);
    return 0;
  }
  if (strcmp(Option, "tsp_version") == 0) {
	 /* do nothing but accept this as a valid option.
	    this allows us to remove this keyword
		from the config file while
		not breaking the client
		when used with older config files
		*/
	  return 0;
  }
  if (strcmp(Option, "server") == 0) {
    Conf->server = strdup(Value);
    return 0;
  }
  if (strcmp(Option, "client_v4") == 0) {
    Conf->client_v4 = strdup(Value);
    return 0;
  }
#ifdef V4V6_SUPPORT  
  if (strcmp(Option, "client_v6") == 0) {
    Conf->client_v6 = strdup(Value);
    return 0;
  }
#endif /* V4V6_SUPPORT  */
  if (strcmp(Option, "userid") == 0) {
    Conf->userid = strdup(Value);
    return 0;
  }
  if (strcmp(Option, "passwd") == 0) {
    Conf->passwd = strdup(Value);
    return 0;
  }
  if (strcmp(Option, "auth_method") == 0) {
    Conf->auth_method = strdup(Value);
    return 0;
  }
  if (strcmp(Option, "host_type") == 0) {
    Conf->host_type = strdup(Value);
    return 0;
  }
  if (strcmp(Option, "template") == 0) {
    Conf->template = strdup(Value);
    return 0;
  }
  if (strcmp(Option, "protocol") == 0) {
    Conf->protocol = strdup(Value);
    return 0;
  }
  if (strcmp(Option, "if_tunnel_v6v4") == 0) {
    Conf->if_tunnel_v6v4 = strdup(Value);
    return 0;
  }
  if (strcmp(Option, "if_tunnel_v6udpv4") == 0) {
    Conf->if_tunnel_v6udpv4 = strdup(Value);
    return 0;
  }
#ifdef V4V6_SUPPORT  
  if (strcmp(Option, "if_tunnel_v4v6") == 0) {
    Conf->if_tunnel_v4v6 = strdup(Value);
    return 0;
  }
#endif /* V4V6_SUPPORT */
  if (strcmp(Option, "tunnel_mode") == 0) {
    if (strcmp(Value, STR_CONFIG_TUNNELMODE_V6ANYV4) == 0) {
      Conf->tunnel_mode = V6ANYV4;
    }
    else if (strcmp(Value, STR_CONFIG_TUNNELMODE_V6V4) == 0) {
      Conf->tunnel_mode = V6V4;
    }
    else if (strcmp(Value, STR_CONFIG_TUNNELMODE_V6UDPV4) == 0) {
      Conf->tunnel_mode = V6UDPV4;
    }
#ifdef V4V6_SUPPORT
    else if (strcmp(Value, STR_CONFIG_TUNNELMODE_V4V6) == 0) {
      Conf->tunnel_mode = V4V6;
    }
#endif /* V4V6_SUPPORT */    
    else {
      Display(LOG_LEVEL_1, ELError, "AddOption", HEX_STR_INVALID_OR_UNSUP_TUNNEL_MODE);
      return 1;
    }
    return 0;
  }
  if (strcmp(Option, "if_source") == 0) {
    Conf->if_source = strdup(Value);
    return 0;
  }
  if (strcmp(Option, "dns_server") == 0) {
    if(VerifyDnsServer(strdup(Value))) {
    	Conf->dns_server = strdup(Value);
	return 0;
    }
    Display(LOG_LEVEL_1, ELError, "AddOption", HEX_STR_DNS_MUST_BE_FQDN);
    return 1;
  }
  if (strcmp(Option, "routing_info") == 0) {
    Conf->routing_info = strdup(Value);
    return 0;
  }
  if (strcmp(Option, "if_prefix") == 0) {
    Conf->if_prefix = strdup(Value);
    return 0;
  }
  if (strcmp(Option, "prefixlen") == 0) {
    Conf->prefixlen = atoi(Value);
    return 0;
  }
  if (strcmp(Option, "retry_delay") == 0) {
     Conf->retry = atoi(Value);
     return 0;
  }
  if (strcmp(Option, "keepalive") == 0) {
    if (strcmp(Value, STR_CONFIG_BOOLEAN_TRUE) == 0) {
      Conf->keepalive = TRUE;
    }
    else if (strcmp(Value, STR_CONFIG_BOOLEAN_FALSE) == 0) {
      Conf->keepalive = FALSE;
    }
    else {
      Display(LOG_LEVEL_1, ELError, "AddOption", HEX_STR_INVALID_VAL_FOR_KEEPALIVE, Value);
      return 1;
    }
    return 0;
  }
  if (strcmp(Option, "keepalive_interval") == 0) {
     Conf->keepalive_interval = atoi(Value);
     return 0;
  }
  if (strcmp(Option, "proxy_client") == 0 ) {
    if (strcmp(Value, STR_CONFIG_BOOLEAN_TRUE) == 0) {
      Conf->proxy_client = TRUE;
    }
    else if (strcmp(Value, STR_CONFIG_BOOLEAN_FALSE) == 0) {
      Conf->proxy_client = FALSE;
    }
    else {
      Display(LOG_LEVEL_1, ELError, "AddOption", HEX_STR_INVALID_VAL_FOR_PROXY_CLIENT, Value);
      return 1;
    }
    return 0;
  }
  if (strcmp(Option, "syslog_level") == 0 ) {
    /* TBD */
    return 0;
  }
  if (strcmp(Option, "syslog_facility") == 0) {
    if (ParseSyslogFacility(Conf, Value) == 0) {
      return 0;	
    }
    else {
      Display(LOG_LEVEL_1, ELError, "AddOption", HEX_STR_INVALID_VAL_FOR_SYSLOG_FACILITY, Value);
      return 1;
	  }
  }
  if (strcmp(Option, "log_filename") == 0) {
	if (strlen(Value) > LOG_FILENAME_MAX_LENGTH) {
		Display(LOG_LEVEL_1, ELError, "AddOption", HEX_STR_LOG_FILENAME_TOO_LONG, LOG_FILENAME_MAX_LENGTH, Value);
		return 1;
	}
	else {
    	Conf->log_filename = strdup(Value);
		return 0;
	}
  }
  if (strcmp(Option, "log_rotation") == 0 ) {
    if (strcmp(Value, STR_CONFIG_BOOLEAN_TRUE) == 0) {
      Conf->log_rotation = TRUE;
    }
    else if (strcmp(Value, STR_CONFIG_BOOLEAN_FALSE) == 0) {
      Conf->log_rotation = FALSE;
    }
    else {
      Display(LOG_LEVEL_1, ELError, "AddOption", HEX_STR_INVALID_VAL_FOR_LOG_ROTATION, Value);
      return 1;
    }
    return 0;
  }
  if (strcmp(Option, "log_rotation_size") == 0) {
    Conf->log_rotation_size = atoi(Value);
	return 0;
  }
  if (strcmp(Option, "log") == 0) {
	return (ParseLogDirective(Conf, Value));
  }
  if (strcmp(Option, "last_server") == 0) {
	if (strlen(Value) > MAX_REDIRECT_LAST_SERVER_LENGTH) {
		Display(LOG_LEVEL_1, ELError, "AddOption", HEX_STR_CFG_LAST_SERVER_TOO_LONG, MAX_REDIRECT_LAST_SERVER_LENGTH);
		return 1;
	}
	else {
		Conf->last_server = strdup(Value);
		return 0;
	}
  }
  if (strcmp(Option, "always_use_same_server") == 0) {
	if (strcmp(Value, STR_CONFIG_BOOLEAN_TRUE) == 0) {
		Conf->always_use_same_server = TRUE;
	}
	else if (strcmp(Value, STR_CONFIG_BOOLEAN_FALSE) == 0) {
		Conf->always_use_same_server = FALSE;
	}
	else {
		Display(LOG_LEVEL_1, ELError, "AddOption", HEX_STR_INVALID_VAL_FOR_AUSS, Value);
		return 1;
	}
	return 0;
  }
  if (strcmp(Option, "broker_list") == 0) {
	if (strlen(Value) > MAX_REDIRECT_BROKER_LIST_LENGTH) {
		Display(LOG_LEVEL_1, ELError, "AddOption", HEX_STR_CFG_BROKER_LIST_TOO_LONG, MAX_REDIRECT_BROKER_LIST_LENGTH);
		return 1;
	}
	else {
		Conf->broker_list = strdup(Value);
		return 0;
	}
  }

  return 0;
}

int
tspReadConfigFile(char *File, tConf *Conf)
{
  FILE *fp;
  int Line = 0, err = 0;
  char *Option, *Value, Buffer[1024], Buffer2[1024], LineBuf[1024];

  log_directive_parsed = 0;

  if ((fp = fopen(File, "r")) == NULL) {
     Display(LOG_LEVEL_1, ELError, "tspReadConfigFile", HEX_STR_ERR_OPEN_CFG_FILE, File);
     return INVALID_CONFIG_FILE;
  }

  while (fgets(Buffer, sizeof(Buffer), fp)) {
    Line++;
    if (*Buffer == '#' || *Buffer == '\r' || *Buffer == '\n')
      continue;

    if (strlen(Buffer) && (Buffer[strlen(Buffer)-1] == '\n' || Buffer[strlen(Buffer)-1] == '\r'))
      Buffer[strlen(Buffer)-1] = '\0';

    if (strlen(Buffer) && (Buffer[strlen(Buffer)-1] == '\n' || Buffer[strlen(Buffer)-1] == '\r'))
      Buffer[strlen(Buffer)-1] = '\0';
     
    if((Option = strtok(Buffer, "=")) != NULL)  {
       if ((Value = strtok(NULL, "")) == NULL) Value = "";
       if (memcmp(Value, "<<", 2)==0) {
         memset(Buffer2, 0, sizeof(Buffer));
         while (fgets(LineBuf, sizeof(LineBuf), fp)) {
            if (*LineBuf == '.') break;
            /* Check for string overflow */
            if (strlen(Buffer2) + strlen(LineBuf) + 1 > sizeof(Buffer2))
               err++;
            else
               strcat(Buffer2, LineBuf);
         }
         Value = Buffer2;
       }
       if (AddOption(Option, Value, Conf)==0)
         continue;
    }

    Display(LOG_LEVEL_1, ELError, "tspReadConfigFile", HEX_STR_ERR_IN_CFG_FILE_LINE, Line);
    err++;
  }

  if (!feof(fp)) {
    err++;
  }

  fclose(fp);

  if (err) {
    Display(LOG_LEVEL_1, ELError, "tspReadConfigFile", HEX_STR_ERRS_IN_CFG_FILE, FileName);
    return(INVALID_CONFIG_FILE);
  }

  return(NO_ERROR);
}

/*
   Initialize with default values, read configuration file and override
   defaults with config file values.
*/
int tspInitialize(int argc, char *argv[], tConf *Conf)
{
  tConf CmdLine;
  char *Templ = "template";
  int status = NO_ERROR;

  Conf->tsp_dir      = NULL;
  Conf->server       = "";
  Conf->userid       = "anonymous";
  Conf->passwd       = "";
  Conf->client_v4    = "auto";
  Conf->client_v6    = "auto";
  Conf->template     = "";
  Conf->dns_server   = "";
  Conf->auth_method  = "any";
  Conf->protocol     = "default_route";
  Conf->tunnel_mode  = V6ANYV4;
  Conf->if_tunnel_v6v4 = "";
  Conf->if_tunnel_v6udpv4 = "";
  Conf->if_tunnel_v4v6 = "";
  Conf->if_source    = "";
  Conf->routing_info = "";
  Conf->if_prefix    = "";
  Conf->host_type    = "host";
  Conf->prefixlen    = 0;
  Conf->retry	     = 30;
  Conf->syslog       = FALSE;
  Conf->keepalive    = TRUE;
  Conf->keepalive_interval = 30;
  Conf->proxy_client = FALSE;
  Conf->syslog_facility = LOG_USER;
  Conf->log_filename = DEFAULT_LOG_FILENAME;
  Conf->log_rotation = TRUE;
  Conf->log_rotation_size = DEFAULT_LOG_ROTATION_SIZE;
  Conf->log_level_stderr = LOG_LEVEL_DISABLED;
  Conf->log_level_console = LOG_LEVEL_DISABLED;
  Conf->log_level_syslog = LOG_LEVEL_DISABLED;
  Conf->log_level_file = LOG_LEVEL_DISABLED;
  Conf->last_server = DEFAULT_REDIRECT_LAST_SERVER_FILE;
  Conf->always_use_same_server = FALSE;
  Conf->broker_list = DEFAULT_REDIRECT_BROKER_LIST_FILE;

  Conf->transport = NET_TOOLS_T_RUDP;  /* RUDP is default value */

  memset(&CmdLine, 0, sizeof(CmdLine));

  if(argc>1) {
	  if ( (status = ParseArguments(argc, argv, &CmdLine)) != NO_ERROR) {
          if (status != NO_ERROR_SHOW_HELP) {
		    Display(LOG_LEVEL_1, ELNotice, "tspInitialize", HEX_STR_ERR_PARSE_CLI_ARGS);		  
          }
		  return status;
	  }
  }

  if((status = tspReadConfigFile(FileName, Conf)) != NO_ERROR) {
	  return status;
  }

  if (log_directive_parsed == 0) {
    Conf->log_level_stderr = DEFAULT_LOG_LEVEL_STDERR;
    Conf->log_level_syslog = DEFAULT_LOG_LEVEL_SYSLOG;
    Conf->log_level_console = DEFAULT_LOG_LEVEL_CONSOLE;
    Conf->log_level_file = DEFAULT_LOG_LEVEL_FILE;
  }

  /* Override the config file with parameters
     from the command line, if any 
  */

  if(CmdLine.if_tunnel_v6v4) Conf->if_tunnel_v6v4 = CmdLine.if_tunnel_v6v4;
  if(CmdLine.if_tunnel_v6udpv4) Conf->if_tunnel_v6udpv4 = CmdLine.if_tunnel_v6udpv4;
#ifdef V4V6_SUPPORT  
  if(CmdLine.if_tunnel_v4v6) Conf->if_tunnel_v4v6 = CmdLine.if_tunnel_v4v6;
#endif /* V4V6_SUPPORT  */
  if(CmdLine.if_source) Conf->if_source = CmdLine.if_source;

  /* Get:

     - The TSP client home directory
     - The directory to run the templates from
 
  */

  if(Conf->tsp_dir) {
     TspHomeDir = Conf->tsp_dir;
     if((ScriptDir = (char *)malloc((size_t)(strlen(Conf->tsp_dir)+strlen(Templ)+2)))==NULL) {
        Display(LOG_LEVEL_3, ELError, "Initialise", HEX_STR_NOT_ENOUGH_MEM);
        return MEMORY_ERROR;
     }
     sprintf(ScriptDir, "%s%c%s", Conf->tsp_dir, DirSeparator, Templ);
  }
  else {
     if((ScriptDir = (char *)malloc((size_t)(strlen(TspHomeDir)+strlen(Templ)+2)))==NULL) {
        Display(LOG_LEVEL_3, ELError, "Initialise", HEX_STR_NOT_ENOUGH_MEM);
        return MEMORY_ERROR;
     }
     sprintf(ScriptDir, "%s%c%s", TspHomeDir, DirSeparator, Templ);
  }
  

#ifdef V4V6_SUPPORT  
  /* Reliable is the default transport protocol.
     For v4v6, the default protocol version is v6, fallback to v4. */
  if (Conf->tunnel_mode == V4V6) {
	  /* be sure config has a v6 server specified */
	  if (NetText2Addr6(Conf->server) != NULL)
		  Conf->transport = NET_TOOLS_T_RUDP6;
	  /* if no v6 server specified, fallback to v4 */
	  else if (NetText2Addr(Conf->server) != NULL)
		  Conf->transport = NET_TOOLS_T_RUDP;
	  else {
		  /* no valid server configured */
		  Display(LOG_LEVEL_3, ELError, "Initialise", HEX_STR_NO_VALID_SERVER);
		  return INVALID_SERVER;
	  }
  }
#endif /* V4V6_SUPPORT */  

  /* For v6v4, the default protocol version is v4, fallback to v6. */
  if (Conf->tunnel_mode == V6V4) {
	  /* be sure config has a v4 server specified */
	  if (NetText2Addr(Conf->server) != NULL)
		  Conf->transport = NET_TOOLS_T_RUDP;
#ifdef V4V6_SUPPORT	  
	  /* if no v4 server specified, fallback to v6 */
	  else if (NetText2Addr6(Conf->server) != NULL)
		  Conf->transport = NET_TOOLS_T_RUDP6;
#endif /* V4V6_SUPPORT */	  
	  else {
		  /* no valid server configured */
		  Display(LOG_LEVEL_3, ELError, "Initialise", HEX_STR_NO_VALID_SERVER);
		  return INVALID_SERVER;
	  }
  }
		  
  /* For v6udpv4, the protocol version MUST BE v4. */
  if (Conf->tunnel_mode == V6UDPV4) {
	  /* be sure config has v4 server specified */
	  if (NetText2Addr(Conf->server) != NULL)
		  Conf->transport = NET_TOOLS_T_RUDP;
	  else {
		  /* cannot fallback on v6udpv4 */
		  Display(LOG_LEVEL_3, ELError, "Initialise", HEX_STR_MODE_V6UDPV4_MUST_CONN_V4);
		  return INVALID_SERVER;
	  }
  }  
  
  return NO_ERROR;
}

int tspUpdateSourceAddr(tConf *Conf, int fd)
{
  int status = NO_ERROR;

  if (Conf->tunnel_mode != V4V6) {
      /* tunnel mode v6*v4, we need source IPv4 address */
      if(IsPresent(Conf->client_v4)) {
	 if(!strcmp(Conf->client_v4, "auto")) { 
	       char addr_str[INET_ADDRSTRLEN+1];

	       /* if current transport is v4, get source address of tsp session. If not, get host address. */
	       if (((Conf->transport == NET_TOOLS_T_RUDP) || (Conf->transport == NET_TOOLS_T_TCP)) &&
		   (tspGetLocalAddress(fd, addr_str, INET_ADDRSTRLEN) != NULL)) {
		       Conf->client_v4 = strdup(addr_str);
		       Display(LOG_LEVEL_3, ELInfo, "tspUpdateSourceAddr", HEX_STR_USING_AS_SOURCE_IPV4, Conf->client_v4);
	       } else {
		       Display(LOG_LEVEL_1, ELError, "tspUpdateSourceAddr", HEX_STR_CANT_AUTO_CLIENT_IPV4_TO_V6);
		       return INVALID_CLIENT_IPV4;
	       }
	       
	 }
      }
  }
#ifdef V4V6_SUPPORT  
  if (Conf->tunnel_mode == V4V6) {
      /* tunnel mode v4v6, we need source IPv6 address. */	   
      if(IsPresent(Conf->client_v6)) {
	     if(!strcmp(Conf->client_v6, "auto")) {
		     char addr_str[INET6_ADDRSTRLEN+1];

		     /* if current transport is v6, get source address of tsp session. If not, get host address. */
		     if (((Conf->transport == NET_TOOLS_T_RUDP6) || (Conf->transport == NET_TOOLS_T_TCP6)) &&
			 (tspGetLocalAddress(fd, addr_str, INET6_ADDRSTRLEN) != NULL)) {
			     Conf->client_v6 = strdup(addr_str);
			     Display(LOG_LEVEL_3, ELInfo, "tspVerifyConfig", HEX_STR_USING_AS_SOURCE_IPV6, Conf->client_v6);
		     } else {
			     Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CANT_AUTO_CLIENT_IPV6_TO_V4);
			     return INVALID_CLIENT_IPV6;
		     }
	     } else {
		     /* remove brackets if any */
		     char addr_str[INET6_ADDRSTRLEN+1];
		     char *ptr;
		     memset(addr_str, 0, sizeof(addr_str));
		     strcpy(addr_str, Conf->client_v6);
		     if ((ptr = strtok(addr_str, "[")) != NULL) {
			     if ((ptr = strtok(ptr, "]")) != NULL) {
				     /* copy back address */
				     strcpy(Conf->client_v6, ptr);
			     }
		     }
	     }     
      }
  }
#endif /* V4V6_SUPPORT */  
  
  return status;
}

/*
   Validate all members of tConf.
*/

int tspVerifyConfig(tConf *Conf)
{
  int status;
# define eq(a,b) (strcmp(a,b) == 0)

  status = 0;  

  if(IsPresent(Conf->client_v4)) {
     if(strcmp(Conf->client_v4, "auto")) { /* If not auto then check if ip address is ok. */
        struct in_addr address;
        unsigned long net;

        net = inet_addr(Conf->client_v4);
		memcpy(&address, &net, sizeof(net));

        if(strcmp(Conf->client_v4, inet_ntoa(address))) {
           Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_BAD_IP_IN_CLIENT_V4, Conf->client_v4);
           goto invalid_config_file;
        }
     }
  }
#ifdef V4V6_SUPPORT
  if(IsPresent(Conf->client_v6)) {
     if(strcmp(Conf->client_v6, "auto")) { /* If not auto then check if ip address is ok. */
        struct in6_addr address;

	if (INET_PTON(AF_INET6, Conf->client_v6, &address) < 0) { 
           Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_BAD_IP_IN_CLIENT_V6, Conf->client_v6);
           goto invalid_config_file;
        }
     }
  }
#endif /* V4V6_SUPPORT */
  
  if(IsPresent(Conf->server)) {
     if(IsAll(ServerString, Conf->server) == 0) {
        Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_BAD_SERVER_NAME);
	goto invalid_config_file;
     }
  } else {
    Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_NEED_SERVER);
    goto invalid_config_file;
  }
  if(IsPresent(Conf->userid)) {
     if(strlen(Conf->userid) > 63) {
        Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_USERID_TOO_LONG);
	goto invalid_config_file;
     } else if(IsAll(DNSLegal, Conf->userid) == 0) {
        Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_BAD_CHARS_USERID);
	goto invalid_config_file;
     }
     if(strcmp(Conf->userid, "anonymous")) {
        if(IsPresent(Conf->passwd)==0) {
           Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_NEED_PASSWD);
	   goto invalid_config_file;
        }
     }
  } else {
    Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_NEED_USERID);
    goto invalid_config_file;
  }

  if(IsPresent(Conf->if_tunnel_v6udpv4)) {
     if(IsAll(DNSLegal, Conf->if_tunnel_v6udpv4) == 0) {
        Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_BAD_CHARS_IF_TUNNEL_V6UDPV4);
         goto invalid_config_file;
     }
  } else {
    if (Conf->tunnel_mode == V6UDPV4) {
      Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_NO_IF_TUNNEL_V6UDPV4);
      goto invalid_config_file;
    }
    else if (Conf->tunnel_mode == V6ANYV4) {
      Display(LOG_LEVEL_2, ELError, "tspVerifyConfig", HEX_STR_CFG_NO_IF_TUNNEL_V6UDPV4_FORCE_V6V4);
      Conf->tunnel_mode = V6V4;
    }
  }
				  
  if(IsPresent(Conf->if_tunnel_v6v4)) {
     if(IsAll(DNSLegal, Conf->if_tunnel_v6v4) == 0) {
        Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_BAD_CHARS_IF_TUNNEL_V6V4);
        goto invalid_config_file;
     }
  } else {
    if (Conf->tunnel_mode == V6V4) {
      Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_NO_IF_TUNNEL_V6V4);
      goto invalid_config_file;
    }
    else if (Conf->tunnel_mode == V6ANYV4) {
      Display(LOG_LEVEL_2, ELError, "tspVerifyConfig", HEX_STR_CFG_NO_IF_TUNNEL_V6V4_FORCE_V6UDPV4);
      Conf->tunnel_mode = V6UDPV4;
    }
  }
  
#ifdef V4V6_SUPPORT  
  if(IsPresent(Conf->if_tunnel_v4v6)) {
     if(IsAll(DNSLegal, Conf->if_tunnel_v4v6) == 0) {
        Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_BAD_CHARS_IF_TUNNEL_V4V6);
        goto invalid_config_file;
     }
  } else {
    if (Conf->tunnel_mode == V4V6) {
      Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_NO_IF_TUNNEL_V4V6);
      goto invalid_config_file;
    }
  }
#endif /* V4V6_SUPPORT */
  
  if ((Conf->proxy_client == TRUE) && (Conf->tunnel_mode == V6UDPV4)) {
      Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_CANT_TUNMODE_AND_PROXY_CLIENT);
      goto invalid_config_file;
  }
  
  if(IsPresent(Conf->template)) {
     if(IsAll(Filename, Conf->template) == 0) {
        Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_BAD_TEMPLATE_NAME);
        goto invalid_config_file;
     }
  } else {
    Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_NEED_TEMPLATE);
    goto invalid_config_file;
  }

  if(IsPresent(Conf->host_type)) {
     if (!(eq(Conf->host_type, "host") ||  eq(Conf->host_type, "router"))) {
       Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_BAD_HOST_TYPE, Conf->host_type);
       goto invalid_config_file;
     }
  }

  if (Conf->prefixlen < 0 || Conf->prefixlen > 64) {
	  Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_CFG_BAD_PREFIX_LEN, Conf->prefixlen);
	  goto invalid_config_file;
  }


  if(IsPresent(Conf->host_type)) {
     if(!(eq(Conf->protocol, "default_route") ||
   	eq(Conf->protocol, "bgp"))) {
       Display(LOG_LEVEL_1,ELError, "tspVerifyConfig", HEX_STR_CFG_BAD_ROUTING_PROTO);
       goto invalid_config_file;
     }
  }
  if(IsPresent(Conf->dns_server)) {
     char *Server;
     char *dns = strdup(Conf->dns_server);
     if (eq(Conf->host_type, "host")) {
       Display(LOG_LEVEL_1,ELError, "tspVerifyConfig", HEX_STR_DNS_DELEG_NOT_SUP_HOSTTYPE_HOST);
       goto invalid_config_file;
     }
     for(Server = strtok(dns, ":");Server; Server = strtok(NULL, ":")) {
        if(gethostbyname(Server) == NULL) {
           Display(LOG_LEVEL_2,ELError, "tspVerifyConfig", HEX_STR_DNS_NOT_RESOLVING, Server);
           goto invalid_config_file;
        }
     }
     free(dns);
  }

  if (Conf->keepalive == TRUE && Conf->keepalive_interval == 0) {
	  Display(LOG_LEVEL_2, ELError, "tspVerifyConfig", HEX_STR_CFG_WARN_KEEPALIVE_TRUE_0);
	  Conf->keepalive = FALSE;
  }

	if (Conf->log_rotation == TRUE && ((Conf->log_rotation_size < LOG_ROTATION_SIZE_MIN) || (Conf->log_rotation_size > LOG_ROTATION_SIZE_MAX))) {
		Display(LOG_LEVEL_1, ELError, "tspVerifyConfig", HEX_STR_INVALID_VAL_FOR_LOG_ROTATION_SIZE, Conf->log_rotation_size);
		goto invalid_config_file;
	}

  return status;

 invalid_config_file:

  return INVALID_CONFIG_FILE;
  
}
