/*
---------------------------------------------------------------------------
 $Id: net_ka.c,v 1.47 2006/11/01 15:16:30 dgregoire Exp $
---------------------------------------------------------------------------
Copyright (c) 2001-2006 Hexago Inc. All rights reserved.

  LICENSE NOTICE: You may use and modify this source code only if you
  have executed a valid license agreement with Hexago Inc. granting
  you the right to do so, the said license agreement governing such
  use and modifications.   Copyright or other intellectual property
  notices are not to be removed from the source code.
---------------------------------------------------------------------------
*/

/*
 This implements a keepalive algorythm.

 Should be able to get a ping socket for all the platforms,
 generate the ICMP6 raw data and process the actual ping
 based on internal values.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define _USES_SYS_TIME_H_
#define _USES_SYS_SOCKET_H_
#define _USES_NETINET_IN_H_
#define _USES_NETINET_IP6_H_
#define _USES_NETINET_ICMP6_H_
#define _USES_NETINET_ICMP_H_
#define _USES_ARPA_INET_H_

#include "platform.h"

#include "net_ka.h"
#include "net_cksm.h"
#include "log.h"	// log levels + display
#include "hex_strings.h"
#include "tsp_lease.h"  /* allocation lease stuff */
#include "errors.h"

static struct NetKeepAlive ka;

static void internal_adjust_major_fuzz(float *);
static void internal_adjust_minor_fuzz(float *);
static void internal_adjust_next_event(struct NetKeepAlive *, int got_reply, int pinged);
static float internal_trim(float, float, float);

static int internal_do_pingout(struct NetKeepAlive *);
static int internal_do_pingin(struct NetKeepAlive *);

static int seqn;

#ifdef BROKEN_WINDOWS_RAWSOCK
// DeviceIOControl ping for windows
HANDLE  Handle;
OVERLAPPED *ol = NULL;
PICMPV6_ER request = NULL;
ICMPV6_EREP reply;
static
void
TranslateSA6toIP6(ADDRESS_IP6 *dst, struct sockaddr_in6 *src)
{
    memcpy(dst, &src->sin6_port, sizeof *dst);
}
#endif


int
NetKeepaliveInit(char *src, char *dst, int maximal_keepalive, int family) {
	struct timeval tv;
#ifdef BROKEN_WINDOWS_RAWSOCK
	if (family == AF_INET6) {
		
	    Handle = CreateFileW(WIN_IPV6_DEVICE_NAME,
			     0,          // desired access
			     FILE_SHARE_READ | FILE_SHARE_WRITE, 
			     NULL,       // security attributes
			     OPEN_EXISTING,
			     FILE_FLAG_OVERLAPPED,          // flags & attributes
			     NULL);      // template file

	    if (Handle == INVALID_HANDLE_VALUE) {
		    Display(LOG_LEVEL_3, ELError, "NetKeepaliveInit", HEX_STR_CANT_CONTACT_V6_DRIVER);
	    return 1;
	    }
	
	} else if (family == AF_INET) {

		if ((Handle = IcmpCreateFile()) == INVALID_HANDLE_VALUE) {
			Display(LOG_LEVEL_3, ELError, "NetKeepaliveInit", HEX_STR_CANT_CREATE_ICMP_HNDL_FILE);
			return 1;
		}
	}
		
#endif

	/* Load the structure with passed in AND
	 * initial values
	 */

	seqn = 0;
	
	if ( strncpy( ka.host, dst, sizeof(ka.host)) == NULL )
		return 1;

	ka.maximal_keepalive = maximal_keepalive;
	ka.current_keepalive = KA_INITIAL_KEEPALIVE;
	ka.major_fuzz = (float) 0.60;
	ka.minor_fuzz = (float) 0.075;

	/* And initialize the next event */
	/* Initial timeout is set to five seconds */

	GETTIMEOFDAY(&tv, NULL);
	ka.next_event.tv_sec = tv.tv_sec + KA_INITIAL_KEEPALIVE;
	ka.next_event.tv_usec = tv.tv_usec;

	/* initialize the random source */

	SRANDOM(tv.tv_usec);

	/* set the family */
	ka.family = family;
	
	/* get a ping socket and a monitor socket */
	if (family == AF_INET6) {
		INET_PTON(AF_INET6, dst, &ka.sa6.sin6_addr);
		ka.sa6.sin6_family = AF_INET6;

		INET_PTON(AF_INET6, src, &ka.sa6_local.sin6_addr);
		ka.sa6_local.sin6_family = AF_INET6;

		ka.keepalive_fd = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
	}
#ifdef V4V6_SUPPORT	
	if (family == AF_INET) {
		INET_PTON(AF_INET, dst, &ka.sa.sin_addr);
		ka.sa.sin_family = AF_INET;

		INET_PTON(AF_INET, src, &ka.sa_local.sin_addr);
		ka.sa_local.sin_family = AF_INET;

		ka.keepalive_fd = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
	}
#endif /* V4V6_SUPPORT */	
	ka.consecutive_timeout = 0;
	ka.still_up = 0;
	ka.doit = 1;
	ka.count = 0;
	ka.got_read = 0;
	ka.got_write = 0;
	ka.pinged = 1;

	/* ok, we are inialized */

	ka.init = 1;
	Display(LOG_LEVEL_3, ELInfo, "NetKeepaliveInit", HEX_STR_KEEPALIVE_INITIALIZED, dst, maximal_keepalive);
	return 0;
}

void
NetKeepaliveDestroy() {
	shutdown(ka.keepalive_fd, SHUT_RDWR);
	close(ka.keepalive_fd);
	memset(&ka, 0, sizeof(ka));
#ifdef BROKEN_WINDOWS_RAWSOCK
	CloseHandle(Handle);
	if (ol != NULL) {
		free(ol);
		ol = NULL;
	}
#endif
}

void
NetKeepaliveGotRead() {
	if (ka.init == 1) {
		ka.got_read = 1;
		//Display(LOG_LEVEL_3, ELNotice, "NetKeepaliveGotRead", HEX_STR_INCOMING_DATA_FROM_TUNNEL);
	}
}

void
NetKeepaliveGotWrite() {
	if (ka.init == 1) {
		ka.got_write = 1;
		//Display(LOG_LEVEL_3, ELNotice, "NetKeepaliveGotWrite", HEX_STR_OUT_DATA_TO_TUNNEL);
	}
}


int
NetKeepaliveV6V4Tunnel(char *src, char *dst, int maximal_keepalive) {
	if (NetKeepaliveInit(src, dst, maximal_keepalive, AF_INET6))
		return 1;
	for (;;) {
		SLEEP(1);
		if ( NetKeepaliveDo() == 2) {
			NetKeepaliveDestroy();
			return KEEPALIVE_TIMEOUT;		// 4 is keepalive timeout. very sketchy. 
		}
	}
	NetKeepaliveDestroy();
	return 0;
}

#ifdef V4V6_SUPPORT
int
NetKeepaliveV4V6Tunnel(char *src, char *dst, int maximal_keepalive, int lifetime) {
	struct timeval tv;
	int expiration = 0;

	if (lifetime != 0) {
		/* allocations lease monitor */
		Display(LOG_LEVEL_3, ELInfo, "NetKeepaliveV4V6Tunnel", HEX_STR_ALLOC_LEASE_MONITOR_ACTIVATED);
		/* get current time */
		GETTIMEOFDAY(&tv, NULL);
		/* calculate expiration time */
		expiration = tv.tv_sec + lifetime;
	}

	/* init keepalive engine */
	if (NetKeepaliveInit(src, dst, maximal_keepalive, AF_INET))
		return 1;
	
	for (;;) {
		SLEEP(1);
		/* check lease expiration */
		if (lifetime != 0)
			if (tspLeaseCheck(expiration) == 1) {
				NetKeepaliveDestroy();
				return LEASE_EXPIRED;
			}
		/* do keepalive */
		if ( NetKeepaliveDo() == 2) {
			NetKeepaliveDestroy();
			return KEEPALIVE_TIMEOUT;		// 4 is keepalive timeout. very sketchy. 
		}
	}
	NetKeepaliveDestroy();
	return 0;
}
#endif /* V4V6_SUPPORT */

/* Should be called with a good resolution 
   to allow for good precision. 

   Every 50 to 100 ms.

   return values :
	0 - everything is OK
	1 - warning - timeout
	2 - fatal - too many timeouts
	3 - not initialised
	*/

int
NetKeepaliveDo() {

	struct timeval tv;
	int doit = 0;		// local evaluation of if we should try to ping.
						// will be flagged only if we hit the next event time
	
	if (ka.init != 1)
		return 3;
	
	if (internal_do_pingin(&ka) != 0)	// read echo replies
		ka.got_read = 1;

	if (ka.got_read == 1) {			// if we did get a read in the last cycle,
		ka.consecutive_timeout = 0;	// we can reset the consecutive timeouts counter.
		ka.count++;					// increment the packet counter
		ka.got_read = 0;			// and reset the read flag until next time
	}

	if (ka.got_write == 1) {	// we did get a write in the last cycle?
		ka.doit = 0;			// then no need to ping out
		ka.got_write = 0;		// and also reset the write flag until next time
	}


	/* do we need to ping out ? */

	GETTIMEOFDAY(&tv, NULL);
	
	/* If a ping is needed,
	   then also take a look
	   at the number of consecutive
	   missreads we have.
	   */

	if (tv.tv_sec == ka.next_event.tv_sec) {
		if (tv.tv_sec >= ka.next_event.tv_usec)
			doit = 1; 
	}

	else if (tv.tv_sec > ka.next_event.tv_sec)
		doit = 1;
	
	
	if (doit) {
	
		if (ka.doit == 0 && ka.count > 0)
			ka.still_up = 1;	/* so if we had both read and writes, we are up? */

		if (ka.consecutive_timeout != 0)	/* if we had timeouts, ping */
			ka.still_up = 0;

		if (ka.count == 2 || ka.count == 1)	/* a count of 2 or 1 probably means */
											/* only the keepalive packet was sent */
			ka.still_up = 0;

		if (ka.doit == 1)					/* no write, do it */
			ka.still_up = 0;

		if ( ka.still_up == 0 )  { 
								// if we havent got any
								// traffic for the last cycle,
								// generate a ping (and thus
								// a reply, and thus - traffic)
								// or none if we are down.
								// a count of 2 will mean only our ping got thru since
								// last time and it means we need to continue
								// pinging
			
			if (internal_do_pingout(&ka) > 0) {
				ka.got_write = 1;	// if it worked, flag the write flag
			}
			ka.consecutive_timeout++;		// we had a timeout of sorts
											// since we had to ping
		}

		ka.doit = 1;		// re-ping unless noted otherwise
		ka.count = 0;		// reset packet counter in between pings
		
		/* adjust internal values */
		/* set the next event time */

		internal_adjust_next_event(&ka, ka.still_up, ka.consecutive_timeout);

		/* adjust the major fuzz factor.
	   it needs to vary +/- 10 - 40% and stay
	   in the .40 - .80 range
	   */
		internal_adjust_major_fuzz(&ka.major_fuzz);

		/* then adjust the minor fuzz factor.
		this one needs to vary +/- 1 - 5%
		staying in the .03 - .07 range
		*/
		internal_adjust_minor_fuzz(&ka.minor_fuzz);

		if (ka.consecutive_timeout == KA_MAXIMAL_CYCLES)
			return 2;

		ka.still_up = 0; // until next time, we are down unless noted otherwise
	}


	/* and return */
	return 0;
}

static 
void
internal_adjust_major_fuzz(float *major_fuzz) {
	int i;
	float f;

	/* here we want a randomness of .40 - .80 */

	i = RANDOM()%40;
    f = (float) i / 100;		/* 0 - 100% */
	f += .40;
	
	*major_fuzz = internal_trim (f, 0.40, 0.80);
}

static
void 
internal_adjust_minor_fuzz(float *minor_fuzz) {
	int i;
	float f;

	/* here we want a randomness of .03 - .07 */

	i = RANDOM()%4;
    f = (float) i / 100;
	f += .03;

	*minor_fuzz = internal_trim (f, 0.03, 0.07);
}

static
void
internal_adjust_next_event(struct NetKeepAlive *nka, int got_reply, int ct) {
	/* if we are in the comfort zone, 
	   apply minor_fuzz.
	   Otherwise, apply major_fuzz

	   If we are in the comfort zone and got no
	   reply, force us out of it.
	*/

	int i;
	struct timeval tv;
	float minor_minimum, major_minimum;

	i = RANDOM()&0x01;      /* positive or negative shift? */

	minor_minimum = KA_INITIAL_KEEPALIVE;
	major_minimum = (float) nka->maximal_keepalive * 0.75;

	/* next_event = current_event * +/- factor */

	if (nka->current_keepalive >= major_minimum) {
			// we are in the comfort zone
			// did we get any reply?
			// if not, bump down outside the zone
			// also ensure that if we
		    // are stuck at either ends of the
		    // spectrum, we get out fast
		if (got_reply == 0 && ct > 1) 
				nka->current_keepalive = major_minimum - 1;
		else {
			float fuzz_factor;
			if (nka->current_keepalive == nka->maximal_keepalive)
				fuzz_factor = 1 - nka->minor_fuzz;
			else if (nka->current_keepalive == major_minimum)
				    fuzz_factor = 1 + nka->minor_fuzz; 
			else fuzz_factor = i ? (1 - nka->minor_fuzz) : (1 + nka->minor_fuzz);
			
			nka->current_keepalive = internal_trim(nka->current_keepalive * fuzz_factor,
				major_minimum, (float) nka->maximal_keepalive);
		}
	}

	/* else we are NOT in the comfort zone */
	if (nka->current_keepalive < major_minimum) {
		/* if we got_reply we go up, otherwise
		   we go down */
		float fuzz_factor;

		fuzz_factor = (ct < 2) ? (1 + nka->major_fuzz) : ( 1 - nka->major_fuzz);
		nka->current_keepalive = internal_trim(nka->current_keepalive * fuzz_factor,
		minor_minimum, (float) nka->maximal_keepalive);
	}

	/* then put the next event in the timeval */

	GETTIMEOFDAY(&tv, NULL);

	nka->next_event.tv_sec  = tv.tv_sec + (int) nka->current_keepalive;
	nka->next_event.tv_usec = ( nka->current_keepalive - (int) nka->current_keepalive ) * 1000000;

	Display(LOG_LEVEL_3, ELNotice, "internal_adjust_next_event", HEX_STR_NEXT_KA_IN, nka->current_keepalive);

	return;
}

static
float
internal_trim(float value, float lower, float higher) {
	if ( value <= lower)
		return lower;
	if (value >= higher)
		return higher;
	return value;
}

/* if this returns 0, no data was written. Otherwise, data was written
*/

static
int
internal_do_pingout(struct NetKeepAlive *nka) {
#ifdef BROKEN_WINDOWS_RAWSOCK
	unsigned long int ret = 0;
	char *SendBuffer;
	LPVOID ReplyBuffer;
	unsigned int SendSize = 16;
	int errorCode, i;
#else
	int ret = 0;
	int len = 0;
	char sendbuf[1500];
#ifdef V4V6_SUPPORT	
	struct icmp *icmp;
#endif /* V4V6_SUPPORT */	
	struct icmp6_hdr *icmp6;
#endif
	
#ifndef BROKEN_WINDOWS_RAWSOCK
	/* compute packet */
#ifdef V4V6_SUPPORT 	
	if (nka->family == AF_INET) {
		/* icmp4 packet */
		icmp = (struct icmp *)sendbuf;
		icmp->icmp_type = ICMP_ECHO;
		icmp->icmp_code = 0;
		icmp->icmp_id = getpid();
		icmp->icmp_seq = seqn++;
		gettimeofday((struct timeval *) icmp->icmp_data, NULL);
		len = 8 /* icmp header */ + 56 /* icmp data */;
		icmp->icmp_cksum = 0;
		icmp->icmp_cksum = in_cksum((u_short *) icmp, len);
		ret = sendto(nka->keepalive_fd, sendbuf, len, 0, (struct sockaddr *) &(nka->sa),
			     sizeof(nka->sa));
	}
#endif /* V4V6_SUPPORT */	
	if (nka->family == AF_INET6) {
		icmp6 = (struct icmp6_hdr *) sendbuf;
		icmp6->icmp6_type = ICMP6_ECHO_REQUEST;
		icmp6->icmp6_code = 0;
		icmp6->icmp6_id = getpid();
		icmp6->icmp6_seq = seqn++;
		gettimeofday((struct timeval *) (icmp6+1), NULL);
		len = 8 /* icmp header */ + 56 /* icmp data */;
		ret = sendto(nka->keepalive_fd, sendbuf, len, 0, (struct sockaddr *) &(nka->sa6),
			     sizeof(nka->sa6));
	}
		
#else

	if (nka->family == AF_INET6) {
	
	    ol = malloc(sizeof(OVERLAPPED));
	    memset(ol, 0, sizeof(OVERLAPPED));

	    if ( (ol->hEvent = CreateEvent(NULL, FALSE, FALSE, "Ping6Event")) == NULL ) {
		    Display(LOG_LEVEL_3, ELError, "NetKeepaliveInit", HEX_STR_CANT_CREATE_EVENT_OBJECT);
	    return 0;
	    }

	    request = malloc(sizeof(ICMPV6_ER) + SendSize * sizeof(char) );
	    memset(request, 0, sizeof(ICMPV6_ER) + SendSize * sizeof(char) );
	    SendBuffer = (char *)(request + 1);

	    TranslateSA6toIP6(&request->src, &ka.sa6_local);
	    TranslateSA6toIP6(&request->dst, &ka.sa6);
	    request->timeout = 10000;	// timeout value in MS

	    for (i = 0; i < SendSize; i++)
	    SendBuffer[i] = 'a' + (i % 26);

	    if ( DeviceIoControl(Handle, IOCTL_ICMPV6_ER,
		      request, sizeof(ICMPV6_ER) + SendSize * sizeof(char),
		      &reply, sizeof(reply), &ret, ol) == FALSE) {
			    if ((errorCode = GetLastError()) == ERROR_IO_PENDING)
				    ret = SendSize;
			    else
				    Display(LOG_LEVEL_3, ELNotice, "internal_do_pingout", HEX_STR_PING_XMIT_FAILED, errorCode);
	    }
	} else if (nka->family == AF_INET) {

	    // Declare and initialize variables
          SendBuffer = malloc(SendSize * sizeof(char));
          memset(SendBuffer, 0, SendSize*sizeof(char));

	    for (i = 0; i < SendSize; i++)
	    SendBuffer[i] = 'a' + (i % 26);
	    
	    ReplyBuffer = (VOID*) malloc(sizeof(ICMP_ECHO_REPLY) + SendSize*sizeof(char));
          memset(ReplyBuffer, 0, sizeof(ICMP_ECHO_REPLY) + SendSize*sizeof(char)); 
	    if ((IcmpSendEcho(Handle, 
	      (nka->sa.sin_addr.s_addr), 
	      SendBuffer, SendSize*sizeof(char), 
	      NULL, ReplyBuffer, 
	      SendSize*sizeof(char) + sizeof(ICMP_ECHO_REPLY),
	      1000)) != 0) {
		    /* data received */
		    ka.consecutive_timeout = 0;
		    ka.count++;
                    free(request);
		    free(ReplyBuffer);
		    return 1;
	    }
	    else {
		    /* error pinging */
		    if ((errorCode = GetLastError()) == ERROR_IO_PENDING) 
			    return 1;
		    else
			    Display(LOG_LEVEL_3, ELNotice, "internal_do_pingout", HEX_STR_PING_XMIT_FAILED, errorCode);
		    return 0;
	    }
 
	} 
#endif
	
	//if (ret > 0)
	//	Display(LOG_LEVEL_3, ELNotice, "internal_do_pingout", HEX_STR_DATA_TO_ICMP6_SOCK, ret);
	return ret;
}
 
/* if this next one returns zero, no ping reply.
   if it returns > 0, ping reply.
   */

static
int
internal_do_pingin(struct NetKeepAlive *nka) {
	unsigned char buffer[2048];

#ifndef BROKEN_WINDOWS_RAWSOCK
	int ret = 0;
	fd_set fs;
	struct timeval tv_sel;
	
	memset(buffer, 0, sizeof(buffer));

	FD_ZERO(&fs);
	FD_SET(nka->keepalive_fd, &fs);
	memset(&tv_sel, 0, sizeof(tv_sel));	// set to zero - imitate polling

	ret = select(nka->keepalive_fd + 1, &fs, NULL, NULL, &tv_sel);	
	if (ret > 0)
		ret = recv(nka->keepalive_fd, buffer, sizeof(buffer), 0);
#else
       unsigned long int ret = 0;
	if (nka->family == AF_INET6) {
	    memset(buffer, 0, sizeof(buffer));
	    if ( ol != NULL && HasOverlappedIoCompleted(ol)) {
		    GetOverlappedResult(Handle, ol, &ret, FALSE);
		    CloseHandle(ol->hEvent);
		    free(ol);
		    ol = NULL;
		    free(request);
		    request = NULL;

		    if (ret > 0 && reply.status != 0)
			    ret = 0;
	    }
	} else {
		/* for AF_INET we get the reply in internal_do_pingout() */
		return 0;
	}
	
#endif
	//if (ret > 0) 
	//	Display(LOG_LEVEL_3, ELNotice, "internal_do_pingin", HEX_STR_ICMP_ECHO_REPLY, ret);
	return ret;
}

